﻿using System;

namespace WirelessSensorGatewayAPI
{
    public abstract class GatewayMessage
    {
        public eMessageType MessageType { get; internal set; }
        public abstract byte[] GetBytes();

        public static byte[] BuildCommandAck(int requestType)
        {
            byte[] packet = new byte[2];

            packet[0] = (byte)eMessageType.CommandAck;
            packet[1] = (byte)requestType;
            
            return packet;
        }

        public static byte[] BuildDataMessage(long sensorID, DateTime utcTime, int status, byte[] sensorData)
        {
            byte[] packet = new byte[sensorData.Length + 11];

            packet[0] = (byte)eMessageType.DataMessage;
            Array.Copy(BitConverter.GetBytes((UInt32)sensorID), 0, packet, 1, 4);
            Array.Copy(BitConverter.GetBytes(Convert.ToUInt32(utcTime.Subtract(new DateTime(2010, 1, 1, 0, 0, 0, DateTimeKind.Utc)).TotalSeconds)), 0, packet, 5, 4);
            packet[9] = (byte)status;
            packet[10] = (byte)sensorData.Length;
            Array.Copy(sensorData, 0, packet, 11, sensorData.Length);

            return packet;
        }

        public static byte[] BuildNetworkStatusRequest()
        {
            byte[] packet = new byte[5];

            packet[0] = (byte)eRequestType.NetworkStatus;
            Array.Copy(BitConverter.GetBytes(Convert.ToUInt32(DateTime.UtcNow.Subtract(new DateTime(2010, 1, 1, 0, 0, 0, DateTimeKind.Utc)).TotalSeconds)), 0, packet, 1, 4);
            
            return packet;
        }

        public static byte[] BuildNetworkStatus(long gatewayID, int channel, bool gatewayActive, int deviceCount)
        {
            byte[] packet = new byte[9];

            packet[0] = (byte)eMessageType.NetworkStatus;
            Array.Copy(BitConverter.GetBytes((UInt32)gatewayID), 0, packet, 1, 4);
            packet[5] = (byte)channel;
            packet[6] = (byte)(gatewayActive ? 0x01 : 0x00);
            Array.Copy(BitConverter.GetBytes((UInt16)deviceCount), 0, packet, 7, 2);
            
            return packet;
        }

        public static byte[] BuildSensorConfigUpdated(long sensorID, bool success)
        {
            byte[] packet = new byte[6];

            packet[0] = (byte)eMessageType.SensorConfigUpdated;
            Array.Copy(BitConverter.GetBytes((UInt32)sensorID), 0, packet, 1, 4);
            packet[5] = (byte)(success ? 0x01 : 0x00);
            
            return packet;
        }

        public static byte[] BuildSensorCommandResponse(long sensorID, int command, bool success)
        {
            byte[] packet = new byte[7];

            packet[0] = (byte)eMessageType.SensorCommandResponse;
            Array.Copy(BitConverter.GetBytes((UInt32)sensorID), 0, packet, 1, 4);
            packet[5] = (byte)command;
            packet[6] = (byte)(success ? 0x01 : 0x00);

            return packet;
        }

        public static byte[] BuildSensorJoined(long sensorID, bool success)
        {
            byte[] packet = new byte[6];

            packet[0] = (byte)eMessageType.SensorJoined;
            Array.Copy(BitConverter.GetBytes((UInt32)sensorID), 0, packet, 1, 4);
            packet[5] = (byte)(success ? 0x01 : 0x00);

            return packet;
        }


        public static byte[] NextPacket(ref byte[] packet)
        {
            int Length;
            switch (packet[0])
            {
                case (byte)eMessageType.CommandAck:
                    Length = 2;
                    break;
                case (byte)eMessageType.DataMessage:
                    int DataLength = packet[10];
                    Length = 11 + DataLength;
                    break;
                case (byte)eMessageType.NetworkStatus:
                    Length = 9;
                    break;
                case (byte)eMessageType.SensorConfigUpdated:
                    Length = 6;
                    break;
                case (byte)eMessageType.SensorCommandResponse:
                    Length = 7;
                    break;
                case (byte)eMessageType.SensorJoined:
                    Length = 6;
                    break;
                default:
                    Length = packet.Length;
                    break;
            }

            byte[] singlePacket = new byte[Length];
            int Remaining = packet.Length - Length;

            Array.Copy(packet, singlePacket, Length);
            if (Remaining > 0)
            {
                byte[] remainingPacket = new byte[Remaining];
                Array.Copy(packet, Length, remainingPacket, 0, Remaining);
                packet = remainingPacket;
            }
            else
            {
                packet = new byte[0];
            }

            return singlePacket;
        }

        public static GatewayMessage ParseGatewayPacket(byte[] packet)
        {
            switch ((eMessageType)packet[0])
            {
                case eMessageType.CommandAck:
                    return new CommandAckMessage(packet);
                case eMessageType.DataMessage:
                    return new DataMessage(packet);
                case eMessageType.NetworkStatus:
                    return new NetworkStatusMessage(packet);
                case eMessageType.SensorConfigUpdated:
                    return new SensorConfigUpdatedMessage(packet);
                case eMessageType.SensorCommandResponse:
                    return new SensorCommandResponseMessage(packet);
                case eMessageType.SensorJoined:
                    return new SensorJoinedMessage(packet);
                default:
                    throw new ArgumentOutOfRangeException("packet",packet[0],"value not an expected message type");
            }
        }

        public static void ParseCommandAck(byte[] packet, out int requestType)
        {
            if (packet[0] != (byte)eMessageType.CommandAck)
            {
                requestType = 0;
                return;
            }

            requestType = packet[1];
        }

        public static void ParseDataMessage(byte[] packet, out long sensorID, out DateTime utcTime, out int status, out byte[] sensorData)
        {
            if(packet[0] != (byte)eMessageType.DataMessage)
            {
                sensorID = 0;
                utcTime = DateTime.MinValue;
                status = 0xFF;
                sensorData = new byte[0];
                return;
            }

            sensorID = BitConverter.ToUInt32(packet, 1);
            utcTime = new DateTime(2010, 1, 1, 0, 0, 0, DateTimeKind.Utc).AddSeconds(BitConverter.ToUInt32(packet, 5));
            status = packet[9];
            int DataLength = packet[10];
            sensorData = new byte[DataLength];
            Array.Copy(packet, 11, sensorData, 0, sensorData.Length);
        }

        public static void ParseNetworkStatus(byte[] packet, out long gatewayID, out int channel, out bool gatewayActive, out int deviceCount)
        {
            if (packet[0] != (byte)eMessageType.NetworkStatus)
            {
                gatewayID = 0;
                channel = 0;
                gatewayActive = false;
                deviceCount = 0;
                return;
            }

            gatewayID = BitConverter.ToUInt32(packet, 1);
            channel = packet[5];
            gatewayActive = packet[6] > 0;
            deviceCount = BitConverter.ToUInt16(packet,7);
        }

        public static void ParseSensorConfigUpdated(byte[] packet, out long sensorID, out bool success)
        {
            if (packet[0] != (byte)eMessageType.SensorConfigUpdated)
            {
                sensorID = 0;
                success = false;
                return;
            }

            sensorID = BitConverter.ToUInt32(packet, 1);
            success = packet[5] == 1;
        }

        public static void ParseSensorCommandResponse(byte[] packet, out long sensorID, out int command, out bool success)
        {
            if (packet[0] != (byte)eMessageType.SensorCommandResponse)
            {
                sensorID = 0;
                command = 0;
                success = false;
                return;
            }

            sensorID = BitConverter.ToUInt32(packet, 1);
            command = packet[5];
            success = packet[6] == 1;
        }

        public static void ParseSensorJoined(byte[] packet, out long sensorID, out bool success)
        {
            if (packet[0] != (byte)eMessageType.SensorJoined)
            {
                sensorID = 0;
                success = false;
                return;
            }

            sensorID = BitConverter.ToUInt32(packet, 1);
            success = packet[5] == 1;
        }
    }

    public class CommandAckMessage : GatewayMessage
    {
        public CommandAckMessage(int requestType)
        {
            MessageType = eMessageType.CommandAck;
            RequestType = requestType;
        }

        public CommandAckMessage(byte[] packet)
        {
            MessageType = eMessageType.CommandAck;
            int OutRequestType;
            GatewayMessage.ParseCommandAck(packet, out OutRequestType);
            RequestType = OutRequestType;
        }

        public int RequestType { get; set; }

        public override byte[] GetBytes()
        {
            return GatewayMessage.BuildCommandAck(RequestType);
        }
    }

    public class DataMessage : GatewayMessage
    {
        public DataMessage(long sensorID, DateTime utcTime, int status, byte[] data)
        {
            MessageType = eMessageType.DataMessage;
            SensorID = sensorID;
            UtcTime = utcTime;
            Status = status;
            Data = data;
        }
        
        public DataMessage(byte[] packet)
        {
            long sensorID;
            DateTime utcTime;
            int status;
            byte[] data;

            GatewayMessage.ParseDataMessage(packet, out sensorID, out utcTime, out status, out data);

            MessageType = eMessageType.DataMessage;
            SensorID = sensorID;
            UtcTime = utcTime;
            Status = status;
            Data = data;
        }

        public long SensorID { get; set; }
        public DateTime UtcTime { get; set; }
        public int Status { get; set; }
        public byte[] Data { get; set; }

        public override byte[] GetBytes()
        {
            return GatewayMessage.BuildDataMessage(SensorID,UtcTime,Status,Data);
        }
    }

    public class NetworkStatusMessage : GatewayMessage
    {
        public NetworkStatusMessage(long gatewayID, int channel, bool active, int deviceCount)
        {
            MessageType = eMessageType.NetworkStatus;
        }

        public NetworkStatusMessage(byte[] packet)
        {
            long gatewayID;
            int channel;
            bool active;
            int deviceCount;

            GatewayMessage.ParseNetworkStatus(packet, out gatewayID, out channel, out active, out deviceCount);

            MessageType = eMessageType.NetworkStatus;
            GatewayID = gatewayID;
            Channel = channel;
            Active = active;
            DeviceCount = DeviceCount;
        }

        public long GatewayID { get; set; }
        public int Channel { get; set; }
        public bool Active { get; set; }
        public int DeviceCount { get; set; }

        public override byte[] GetBytes()
        {
            return BuildNetworkStatus(GatewayID, Channel, Active, DeviceCount);
        }
    }

    public class SensorConfigUpdatedMessage : GatewayMessage
    {
        public SensorConfigUpdatedMessage(long sensorID, bool success)
        {
            MessageType = eMessageType.SensorConfigUpdated;
            SensorID = sensorID;
            Success = success;
        }
        public SensorConfigUpdatedMessage(byte[] packet)
        {
            long sensorID;
            bool success;

            GatewayMessage.ParseSensorConfigUpdated(packet, out sensorID, out success);

            MessageType = eMessageType.SensorConfigUpdated;
            SensorID = sensorID;
            Success = success;
        }

        public long SensorID { get; set; }
        public bool Success { get; set; }

        public override byte[] GetBytes()
        {
            return GatewayMessage.BuildSensorConfigUpdated(SensorID, Success);
        }
    }

    public class SensorCommandResponseMessage : GatewayMessage
    {
        public SensorCommandResponseMessage(long sensorID, int command, bool success)
        {
            MessageType = eMessageType.SensorCommandResponse;
            SensorID = sensorID;
            Command = command;
            Success = success;
        }
        public SensorCommandResponseMessage(byte[] packet)
        {
            long sensorID;
            int command;
            bool success;

            GatewayMessage.ParseSensorCommandResponse(packet, out sensorID, out command, out success);

            MessageType = eMessageType.SensorCommandResponse;
            SensorID = sensorID;
            Command = command;
            Success = success;
        }

        public long SensorID { get; set; }
        public int Command { get; set; }
        public bool Success { get; set; }

        public override byte[] GetBytes()
        {
            return GatewayMessage.BuildSensorCommandResponse(SensorID, Command, Success);
        }
    }

    public class SensorJoinedMessage : GatewayMessage
    {
        public SensorJoinedMessage(long sensorID, bool success)
        {
            MessageType = eMessageType.SensorJoined;
            SensorID = sensorID;
            Success = success;
        }
        public SensorJoinedMessage(byte[] packet)
        {
            long sensorID;
            bool success;

            GatewayMessage.ParseSensorJoined(packet, out sensorID, out success);

            MessageType = eMessageType.SensorJoined;
            SensorID = sensorID;
            Success = success;
        }

        public long SensorID { get; set; }
        public bool Success { get; set; }

        public override byte[] GetBytes()
        {
            return GatewayMessage.BuildSensorConfigUpdated(SensorID, Success);
        }
    }
}
